# coding: utf-8
# 2021/8/6 @ zengxiaonan

import pytest
from EduKTM import AKT


@pytest.mark.parametrize("maxgradnorm", [-1, 1])
@pytest.mark.parametrize("separate_qa", [False, True])
@pytest.mark.parametrize("kq_same", [0, 1])
def test_train(data, conf, tmp_path, maxgradnorm, separate_qa, kq_same):
    n_question, n_pid, batch_size = conf
    n_blocks = 1
    d_model = 32
    dropout = 0.05
    l2 = 1e-5

    akt = AKT(n_question, n_pid, n_blocks, d_model, dropout, kq_same, l2, batch_size, maxgradnorm, separate_qa)
    akt.train(data, test_data=data, epoch=2)
    filepath = tmp_path / "dkt+.params"
    akt.save(filepath)
    akt.load(filepath)



# coding: utf-8
# 2021/8/6 @ zengxiaonan
import random
import numpy as np
import pytest


@pytest.fixture(scope="package", params=[0, 100])
def conf(request):
    batch_size = 32
    n_question = 10
    n_pid = request.param
    return n_question, n_pid, batch_size


@pytest.fixture(scope="package", params=[0, 10])
def data(conf, request):
    n_question, n_pid, batch_size = conf
    batch_size += request.param
    seqlen = 10

    q = [
        [random.randint(1, n_question) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    a = [
        [random.randint(0, 1) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    qa = [
        [a[i][j] * n_question + q_id for j, q_id in enumerate(q_seq)]
        for i, q_seq in enumerate(q)
    ]
    if n_pid > 0:
        p = [
            [random.randint(1, n_pid) for _ in range(seqlen)]
            for _ in range(batch_size)
        ]
    else:
        p = []
    data = (np.array(q), np.array(qa), np.array(p))

    return data



# coding: utf-8
# 2021/5/26 @ tongshiwei
# These codes are modified from gluonnlp

import math
import numpy as np
import warnings


def _match_bucket_keys(bucket_keys, seq_lengths):
    bucket_key_npy = np.array(bucket_keys, dtype=np.int32)
    bucket_sample_ids = [list() for _ in range(len(bucket_keys))]
    batch_size = 10000
    bucket_key_npy = bucket_key_npy.reshape((1,) + bucket_key_npy.shape)
    for begin in range(0, len(seq_lengths), batch_size):
        end = min(begin + batch_size, len(seq_lengths))
        diff = bucket_key_npy - np.expand_dims(seq_lengths[begin:end], axis=1)
        if diff.ndim == 3:
            is_valid_bucket = np.prod(diff >= 0, axis=2)
            pad_val = np.sum(diff, axis=2)
        else:
            is_valid_bucket = diff >= 0
            pad_val = diff
        seq_ids_not_found = np.nonzero(is_valid_bucket.sum(axis=1) == 0)[0].tolist()
        masked_pad_val = np.ma.array(pad_val, mask=1 - is_valid_bucket)
        batch_bucket_id = masked_pad_val.argmin(axis=1).tolist()
        if len(seq_ids_not_found) > 0:
            raise ValueError('Find elements in seq_lengths that cannot fit in the '
                             'given buckets, seq_length=%s, bucket_keys=%s. ' \
                             'You must increase the bucket size.'
                             % (str(seq_lengths[seq_ids_not_found]), str(bucket_keys)))
        for i, bucket_id in enumerate(batch_bucket_id):
            bucket_sample_ids[bucket_id].append(i + begin)
    return bucket_sample_ids


def _bucket_stats(bucket_sample_ids, seq_lengths):
    bucket_average_lengths = []
    bucket_length_stds = []
    for sample_ids in bucket_sample_ids:
        if len(sample_ids) > 0:
            lengths = seq_lengths[sample_ids]
            bucket_average_lengths.append(np.mean(lengths))
            bucket_length_stds.append(np.std(lengths))
        else:
            bucket_average_lengths.append(0)
            bucket_length_stds.append(0)
    return (bucket_average_lengths, bucket_length_stds)


class BucketScheme:
    r"""Base class for generating bucket keys."""

    def __call__(self, max_lengths, min_lengths, num_buckets):
        """Generate bucket keys based on the lengths of sequences and number of buckets.

        Parameters
        ----------
        max_lengths : int or list of int
            Maximum of lengths of sequences.
        min_lengths : int or list of int
            Minimum of lengths of sequences.
        num_buckets : int
            Number of buckets

        Returns
        -------
        bucket_keys : list of int
            A list including the keys of the buckets.
        """
        raise NotImplementedError


class ConstWidthBucket(BucketScheme):
    r"""Buckets with constant width."""

    def __call__(self, max_lengths, min_lengths, num_buckets):
        r"""This generate bucket keys given that all the buckets have the same width.

        Parameters
        ----------
        max_lengths : int or list of int
            Maximum of lengths of sequences.
        min_lengths : int or list of int
            Minimum of lengths of sequences.
        num_buckets : int
            Number of buckets

        Returns
        -------
        bucket_keys : list of int
            A list including the keys of the buckets.
        """
        if isinstance(max_lengths, list):
            bucket_width_l = [max((1 + max_len - min_len) // num_buckets, 1)
                              for max_len, min_len in
                              zip(max_lengths, min_lengths)]
            bucket_keys = \
                [tuple(max(max_len - i * width, min_len) for max_len, min_len, width in
                       zip(max_lengths, min_lengths, bucket_width_l))
                 for i in range(num_buckets)]
        else:
            bucket_width = max((1 + max_lengths - min_lengths) // num_buckets, 1)
            bucket_keys = [max(max_lengths - i * bucket_width, min_lengths)
                           for i in range(num_buckets)]
        return bucket_keys


class LinearWidthBucket(BucketScheme):
    r""" Buckets with linearly increasing width:
    :math:`w_i = \alpha * i + 1` for all :math:`i \geq 1`.
    """

    def __call__(self, max_lengths, min_lengths, num_buckets):
        r"""This function generates bucket keys with linearly increasing bucket width:

        Parameters
        ----------
        max_lengths : int or list of int
            Maximum of lengths of sequences.
        min_lengths : int or list of int
            Minimum of lengths of sequences.
        num_buckets : int
            Number of buckets

        Returns
        -------
        bucket_keys : list of int
            A list including the keys of the buckets.
        """
        if isinstance(max_lengths, list):
            alpha_l = [2 * float(max_len - min_len - num_buckets)
                       / (num_buckets * (num_buckets + 1))
                       for max_len, min_len in
                       zip(max_lengths, min_lengths)]
            bucket_keys = \
                [tuple(int(round(min_len + alpha * (((i + 1) * (i + 2)) / 2) + i + 1))
                       for min_len, alpha in zip(min_lengths, alpha_l))
                 for i in range(num_buckets)]
            bucket_keys[-1] = tuple(max(max_bucket_key, max_len)
                                    for max_bucket_key, max_len
                                    in zip(bucket_keys[-1], max_lengths))
        else:
            alpha = 2 * float(max_lengths - min_lengths - num_buckets) \
                    / (num_buckets * (num_buckets + 1))
            bucket_keys = [int(round(min_lengths + alpha * (((i + 1) * (i + 2)) / 2) + i + 1))
                           for i in range(num_buckets)]
            bucket_keys[-1] = max(bucket_keys[-1], max_lengths)
        return bucket_keys


class ExpWidthBucket(BucketScheme):
    r""" Buckets with exponentially increasing width:
    :math:`w_i = bucket\_len\_step * w_{i-1}` for all :math:`i \geq 2`.

    Parameters
    ----------
    bucket_len_step : float, default 1.1
        This is the increasing factor for the bucket width.
    """

    def __init__(self, bucket_len_step=1.1):
        self.bucket_len_step = bucket_len_step

    def __call__(self, max_lengths, min_lengths, num_buckets):
        r"""This function generates bucket keys exponentially increasing bucket width.

        Parameters
        ----------
        max_lengths : int or list of int
            Maximum of lengths of sequences.
        min_lengths : int or list of int
            Minimum of lengths of sequences.
        num_buckets : int
            Number of buckets

        Returns
        -------
        bucket_keys : list of int
            A list including the keys of the buckets.
        """
        if isinstance(max_lengths, list):
            initial_width_l = [
                (max_len - min_len) * (self.bucket_len_step - 1)
                / (math.pow(self.bucket_len_step, num_buckets) - 1)
                for max_len, min_len in
                zip(max_lengths, min_lengths)]
            bucket_keys = \
                [tuple(
                    int(round(min_len + initial_width * (math.pow(self.bucket_len_step, i + 1) - 1)
                              / (self.bucket_len_step - 1)))
                    for min_len, initial_width in zip(min_lengths, initial_width_l))
                    for i in range(num_buckets)]
            bucket_keys[-1] = tuple(max(max_bucket_key, max_len)
                                    for max_bucket_key, max_len
                                    in zip(bucket_keys[-1], max_lengths))
        else:
            initial_width = (max_lengths - min_lengths) * (self.bucket_len_step - 1) \
                            / (math.pow(self.bucket_len_step, num_buckets) - 1)
            bucket_keys = [
                int(round(min_lengths + initial_width * (math.pow(self.bucket_len_step, i + 1) - 1)
                          / (self.bucket_len_step - 1)))
                for i in range(num_buckets)]
            bucket_keys[-1] = max(bucket_keys[-1], max_lengths)
        return bucket_keys


class Sampler(object):
    """Base class for samplers.

    All samplers should subclass `Sampler` and define `__iter__` and `__len__`
    methods.
    """

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError


class FixedBucketSampler(Sampler):
    r"""Assign each data sample to a fixed bucket based on its length.
    The bucket keys are either given or generated from the input sequence lengths.

    Parameters
    ----------
    lengths : list of int or list of tuple/list of int
        The length of the sequences in the input data sample.
    batch_size : int
        The batch size of the sampler.
    num_buckets : int or None, default 10
        The number of buckets. This will not be used if bucket_keys is set.
    bucket_keys : None or list of int or list of tuple, default None
        The keys that will be used to create the buckets. It should usually be the lengths of the
        sequences. If it is None, the bucket_keys will be generated based on the maximum
        lengths of the data.
    ratio : float, default 0
        Ratio to scale up the batch size of smaller buckets.
        Assume the :math:`i` th key is :math:`K_i` ,
        the default batch size is :math:`B` , the ratio to scale the batch size is
        :math:`\alpha` and
        the batch size corresponds to the :math:`i` th bucket is :math:`B_i` . We have:

        .. math::

            B_i = \max(\alpha B \times \frac{\max_j sum(K_j)}{sum(K_i)}, B)

        Thus, setting this to a value larger than 0, like 0.5, will scale up the batch size of the
        smaller buckets.
    shuffle : bool, default False
        Whether to shuffle the batches.
    use_average_length : bool, default False
        False: each batch contains batch_size sequences, number of sequence elements varies.
        True: each batch contains batch_size elements, number of sequences varies. In this case,
        ratio option is ignored.
    num_shards : int, default 0
        If num_shards > 0, the sampled batch is split into num_shards smaller batches.
        The output will have structure of list(list(int)).
        If num_shards = 0, the output will have structure of list(int).
        This is useful in multi-gpu training and can potentially reduce the number of paddings.
        In general, it is set to the number of gpus.
    bucket_scheme : BucketScheme, default ConstWidthBucket
        It is used to generate bucket keys. It supports:
        ConstWidthBucket: all the buckets have the same width
        LinearWidthBucket: the width of ith  bucket follows :math:`w_i = \alpha * i + 1`
        ExpWidthBucket: the width of ith bucket follows
        :math:`w_i` = bucket_len_step :math:`* w_{i-1}`
    Examples
    --------
    >>> lengths = [np.random.randint(1, 100) for _ in range(1000)]
    >>> sampler = FixedBucketSampler(lengths, 8, ratio=0.5)
    >>> print(sampler.stats())
    FixedBucketSampler:
    -etc-
    """

    def __init__(self, lengths, batch_size, num_buckets=10, bucket_keys=None,
                 ratio=0, shuffle=False, use_average_length=False, num_shards=0,
                 bucket_scheme=ConstWidthBucket()):
        assert len(lengths) > 0, 'FixedBucketSampler does not support empty lengths.'
        assert batch_size > 0, 'Batch size must be larger than 0.'
        assert ratio >= 0, 'batch size scaling ratio cannot be negative.'
        self._batch_size = batch_size
        self._ratio = ratio
        self._lengths = np.array(lengths, dtype=np.int32)
        if self._lengths.ndim == 1:
            self._single_element = True
            attr_num = 1
        else:
            assert self._lengths.ndim == 2, \
                'Elements in lengths must be either int or tuple/list of int. ' \
                'Received lengths=%s' % str(lengths)
            self._single_element = False
            attr_num = self._lengths.shape[1]
        self._shuffle = shuffle
        self._num_shards = num_shards
        self._bucket_scheme = bucket_scheme
        max_lengths = self._lengths.max(axis=0)
        min_lengths = self._lengths.min(axis=0)
        if self._single_element:
            assert min_lengths > 0, 'Sequence lengths must all be larger than 0.'
        else:
            for _, ele in enumerate(min_lengths):
                assert ele > 0, 'Sequence lengths must all be larger than 0.'
        # Generate the buckets
        if bucket_keys is None:
            assert num_buckets > 0, 'num_buckets must be set when bucket_keys is None. Received ' \
                                    'num_buckets=%d' % num_buckets
            bucket_keys = bucket_scheme(max_lengths, min_lengths, num_buckets)
        else:
            if num_buckets is not None:
                warnings.warn('num_buckets will not be used if bucket_keys is not None. '
                              'bucket_keys=%s, num_buckets=%d' % (str(bucket_keys), num_buckets))
            assert len(bucket_keys) > 0
            if self._single_element:
                assert isinstance(bucket_keys[0], int)
            else:
                assert isinstance(bucket_keys[0], tuple)
                assert len(bucket_keys[0]) == attr_num
        bucket_keys = sorted(set(bucket_keys))
        # Assign instances to buckets
        bucket_sample_ids = _match_bucket_keys(bucket_keys, self._lengths)
        unused_bucket_keys = [key for key, sample_ids in zip(bucket_keys, bucket_sample_ids)
                              if len(sample_ids) == 0]
        if len(unused_bucket_keys) > 0:
            warnings.warn('Some buckets are empty and will be removed. Unused bucket keys=%s' %
                          str(unused_bucket_keys))
        # Remove empty buckets
        self._bucket_keys = [key for key, sample_ids in zip(bucket_keys, bucket_sample_ids)
                             if len(sample_ids) > 0]

        self._bucket_sample_ids = [sample_ids for sample_ids in bucket_sample_ids
                                   if len(sample_ids) > 0]
        if not use_average_length:
            scale_up_keys = [key if self._single_element else sum(key) for key
                             in self._bucket_keys]
            max_scale_up_key = max(scale_up_keys)
            self._bucket_batch_sizes = [max(int(max_scale_up_key / float(scale_up_key)
                                                * self._ratio * batch_size), batch_size)
                                        for scale_up_key in scale_up_keys]
        else:
            if ratio > 0.:
                warnings.warn('ratio=%f is ignored when use_average_length is True' % self._ratio)
            bucket_average_lengths, bucket_length_stds = _bucket_stats(self._bucket_sample_ids,
                                                                       self._lengths)
            self._bucket_batch_sizes = [max(int(batch_size / (average_length + length_std)), 1)
                                        for average_length, length_std
                                        in zip(bucket_average_lengths, bucket_length_stds)]
        self._batch_infos = []
        for bucket_id, sample_ids, bucket_batch_size in \
                zip(range(len(self._bucket_keys) - 1, -1, -1),
                    self._bucket_sample_ids[::-1],
                    self._bucket_batch_sizes[::-1]):
            for i in range(0, len(sample_ids), bucket_batch_size):
                self._batch_infos.append((bucket_id, i))

        if self._num_shards > 0:
            self._sampler_size = int(math.ceil(len(self._batch_infos) / float(self._num_shards)))
        else:
            self._sampler_size = len(self._batch_infos)

    def __iter__(self):
        if self._shuffle:
            np.random.shuffle(self._batch_infos)
            for bucket_id in range(len(self._bucket_keys)):
                np.random.shuffle(self._bucket_sample_ids[bucket_id])

        if self._num_shards > 0:
            for batch_idx in range(0, len(self._batch_infos), self._num_shards):
                if batch_idx + self._num_shards > len(self._batch_infos):
                    batch_idx = len(self._batch_infos) - self._num_shards
                batch = self._batch_infos[batch_idx: batch_idx + self._num_shards]
                bucket_ids, batch_begins = list(zip(*batch))
                batch_sizes = [self._bucket_batch_sizes[bucket_id] for bucket_id in bucket_ids]
                batch_ends = [min(batch_begin + batch_size,
                                  len(self._bucket_sample_ids[bucket_id]))
                              for bucket_id, batch_begin, batch_size in zip(bucket_ids,
                                                                            batch_begins,
                                                                            batch_sizes)]
                yield [self._bucket_sample_ids[bucket_id][batch_begin:batch_end]
                       for bucket_id, batch_begin, batch_end in zip(bucket_ids,
                                                                    batch_begins,
                                                                    batch_ends)]
        else:
            for bucket_id, batch_begin in self._batch_infos:
                batch_size = self._bucket_batch_sizes[bucket_id]
                batch_end = min(batch_begin + batch_size, len(self._bucket_sample_ids[bucket_id]))
                yield self._bucket_sample_ids[bucket_id][batch_begin:batch_end]

    def __len__(self):
        return self._sampler_size

    def stats(self):
        """Return a string representing the statistics of the bucketing sampler.

        Returns
        -------
        ret : str
            String representing the statistics of the buckets.
        """
        ret = '{name}:\n' \
              '  sample_num={sample_num}, batch_num={batch_num}\n' \
              '  key={bucket_keys}\n' \
              '  cnt={bucket_counts}\n' \
              '  batch_size={bucket_batch_sizes}' \
            .format(name=self.__class__.__name__,
                    sample_num=len(self._lengths),
                    batch_num=len(self._batch_infos),
                    bucket_keys=self._bucket_keys,
                    bucket_counts=[len(sample_ids) for sample_ids in self._bucket_sample_ids],
                    bucket_batch_sizes=self._bucket_batch_sizes)
        return ret


class SortedBucketSampler(Sampler):
    r"""Batches are sampled from sorted buckets of data.

    First, partition data in buckets of size `batch_size * mult`.
    Each bucket contains `batch_size * mult` elements. The samples inside each bucket are sorted
    based on sort_key and then batched.

    Parameters
    ----------
    sort_keys : list-like object
        The keys to sort the samples.
    batch_size : int
        Batch size of the sampler.
    mult : int or float, default 100
        The multiplier to determine the bucket size. Each bucket will have size `mult * batch_size`.
    reverse : bool, default True
        Whether to sort in descending order.
    shuffle : bool, default False
        Whether to shuffle the data.

    Examples
    --------
    >>> lengths = [np.random.randint(1, 1000) for _ in range(1000)]
    >>> sampler = SortedBucketSampler(lengths, 16)
    >>> # The sequence lengths within the batch will be sorted
    >>> for i, indices in enumerate(sampler):
    ...     if i == 0:
    ...         print([lengths[ele] for ele in indices])
    [-etc-]
    """

    def __init__(self, sort_keys, batch_size, mult=100, reverse=True, shuffle=False):
        assert len(sort_keys) > 0
        assert batch_size > 0
        assert mult >= 1, 'Bucket size multiplier must be larger than 1'
        self._sort_keys = sort_keys
        self._batch_size = batch_size
        self._mult = mult
        self._total_sample_num = len(self._sort_keys)
        self._reverse = reverse
        self._shuffle = shuffle

    def __iter__(self):
        if self._shuffle:
            sample_ids = np.random.permutation(self._total_sample_num)
        else:
            sample_ids = list(range(self._total_sample_num))
        bucket_size = int(self._mult * self._batch_size)
        for bucket_begin in range(0, self._total_sample_num, bucket_size):
            bucket_end = min(bucket_begin + bucket_size, self._total_sample_num)
            sorted_sample_ids = sorted(sample_ids[bucket_begin:bucket_end],
                                       key=lambda i: self._sort_keys[i], reverse=self._reverse)
            batch_begins = list(range(0, len(sorted_sample_ids), self._batch_size))
            if self._shuffle:
                np.random.shuffle(batch_begins)
            for batch_begin in batch_begins:
                batch_end = min(batch_begin + self._batch_size, len(sorted_sample_ids))
                yield sorted_sample_ids[batch_begin:batch_end]

    def __len__(self):
        return (len(self._sort_keys) + self._batch_size - 1) // self._batch_size



# coding: utf-8
# 2021/8/21 @ zengxiaonan
import random
import numpy as np
import pytest


@pytest.fixture(scope="package")
def conf():
    batch_size = 16
    n_at = 64
    n_it = 32
    n_question = 8
    n_exercise = 32

    q_matrix = np.zeros((n_exercise + 1, n_question + 1)) + 0.3
    for row_id in range(n_exercise + 1):
        rand_idx = random.randint(1, n_question)
        q_matrix[row_id][rand_idx] = 1
    return n_at, n_it, n_question, n_exercise, q_matrix, batch_size


@pytest.fixture(scope="package", params=[0, 8])
def data(conf, request):
    n_at, n_it, _, n_exercise, _, batch_size = conf
    batch_size += request.param
    seqlen = 10

    a = [
        [random.randint(0, 1) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    e = [
        [random.randint(1, n_exercise) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    it = [
        [random.randint(1, n_it) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    at = [
        [random.randint(1, n_at) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    data = (np.array(a), np.array(e), np.array(it), np.array(at))

    return data



# coding: utf-8
# 2022/3/18 @ ouyangjie


import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import math
from sklearn import metrics
from tqdm import tqdm
from EduKTM import KTM


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Cell(nn.Module):
    def __init__(self, memory_size, memory_state_dim):
        super(Cell, self).__init__()
        self.memory_size = memory_size
        self.memory_state_dim = memory_state_dim

    def addressing(self, control_input, memory):
        """
        Parameters
        ----------
        control_input: tensor
            embedding vector of input exercise, shape = (batch_size, control_state_dim)
        memory: tensor
            key memory, shape = (memory_size, memory_state_dim)

        Returns
        -------
        correlation_weight: tensor
            correlation weight, shape = (batch_size, memory_size)
        """
        similarity_score = torch.matmul(control_input, torch.t(memory))
        correlation_weight = F.softmax(similarity_score, dim=1)  # Shape: (batch_size, memory_size)
        return correlation_weight

    def read(self, memory, read_weight):
        """
        Parameters
        ----------
        memory: tensor
            value memory, shape = (batch_size, memory_size, memory_state_dim)
        read_weight: tensor
            correlation weight, shape = (batch_size, memory_size)

        Returns
        -------
        read_content: tensor
            read content, shape = (batch_size, memory_size)
        """
        read_weight = read_weight.view(-1, 1)
        memory = memory.view(-1, self.memory_state_dim)
        rc = torch.mul(read_weight, memory)
        read_content = rc.view(-1, self.memory_size, self.memory_state_dim)
        read_content = torch.sum(read_content, dim=1)
        return read_content


class WriteCell(Cell):
    def __init__(self, memory_size, memory_state_dim):
        super(WriteCell, self).__init__(memory_size, memory_state_dim)
        self.erase = torch.nn.Linear(memory_state_dim, memory_state_dim, bias=True)
        self.add = torch.nn.Linear(memory_state_dim, memory_state_dim, bias=True)
        nn.init.kaiming_normal_(self.erase.weight)
        nn.init.kaiming_normal_(self.add.weight)
        nn.init.constant_(self.erase.bias, 0)
        nn.init.constant_(self.add.bias, 0)

    def write(self, control_input, memory, write_weight):
        """
        Parameters
        ----------
        control_input: tensor
            embedding vector of input exercise and students' answer, shape = (batch_size, control_state_dim)
        memory: tensor
            value memory, shape = (batch_size, memory_size, memory_state_dim)
        read_weight: tensor
            correlation weight, shape = (batch_size, memory_size)

        Returns
        -------
        new_memory: tensor
            updated value memory, shape = (batch_size, memory_size, memory_state_dim)
        """
        erase_signal = torch.sigmoid(self.erase(control_input))
        add_signal = torch.tanh(self.add(control_input))
        erase_reshape = erase_signal.view(-1, 1, self.memory_state_dim)
        add_reshape = add_signal.view(-1, 1, self.memory_state_dim)
        write_weight_reshape = write_weight.view(-1, self.memory_size, 1)
        erase_mult = torch.mul(erase_reshape, write_weight_reshape)
        add_mul = torch.mul(add_reshape, write_weight_reshape)
        new_memory = memory * (1 - erase_mult) + add_mul
        return new_memory


class DKVMNCell(nn.Module):
    def __init__(self, memory_size, key_memory_state_dim, value_memory_state_dim, init_key_memory):
        super(DKVMNCell, self).__init__()
        """
        Parameters
        ----------
        memory_size: int
            size of memory
        key_memory_state_dim: int
            dimension of key memory
        value_memory_state_dim:  int
            dimension of value memory
        init_key_memory: tensor
            intial key memory
        """
        self.memory_size = memory_size
        self.key_memory_state_dim = key_memory_state_dim
        self.value_memory_state_dim = value_memory_state_dim

        self.key_head = Cell(memory_size=self.memory_size, memory_state_dim=self.key_memory_state_dim)
        self.value_head = WriteCell(memory_size=self.memory_size, memory_state_dim=self.value_memory_state_dim)

        self.key_memory = init_key_memory
        self.value_memory = None

    def init_value_memory(self, value_memory):
        self.value_memory = value_memory

    def attention(self, control_input):
        correlation_weight = self.key_head.addressing(control_input=control_input, memory=self.key_memory)
        return correlation_weight

    def read(self, read_weight):
        read_content = self.value_head.read(memory=self.value_memory, read_weight=read_weight)
        return read_content

    def write(self, write_weight, control_input):
        value_memory = self.value_head.write(control_input=control_input,
                                             memory=self.value_memory,
                                             write_weight=write_weight)
        self.value_memory = nn.Parameter(value_memory.data)

        return self.value_memory


class Net(nn.Module):
    def __init__(self, n_question, batch_size, key_embedding_dim, value_embedding_dim,
                 memory_size, key_memory_state_dim, value_memory_state_dim, final_fc_dim, student_num=None):
        super(Net, self).__init__()
        self.n_question = n_question
        self.batch_size = batch_size
        self.key_embedding_dim = key_embedding_dim
        self.value_embedding_dim = value_embedding_dim
        self.memory_size = memory_size
        self.key_memory_state_dim = key_memory_state_dim
        self.value_memory_state_dim = value_memory_state_dim
        self.final_fc_dim = final_fc_dim
        self.student_num = student_num

        self.input_embed_linear = nn.Linear(self.key_embedding_dim, self.final_fc_dim, bias=True)
        self.read_embed_linear = nn.Linear(self.value_memory_state_dim + self.final_fc_dim,
                                           self.final_fc_dim, bias=True)
        self.predict_linear = nn.Linear(self.final_fc_dim, 1, bias=True)
        self.init_key_memory = nn.Parameter(torch.randn(self.memory_size, self.key_memory_state_dim))
        nn.init.kaiming_normal_(self.init_key_memory)
        self.init_value_memory = nn.Parameter(torch.randn(self.memory_size, self.value_memory_state_dim))
        nn.init.kaiming_normal_(self.init_value_memory)

        self.mem = DKVMNCell(memory_size=self.memory_size, key_memory_state_dim=self.key_memory_state_dim,
                             value_memory_state_dim=self.value_memory_state_dim, init_key_memory=self.init_key_memory)

        value_memory = nn.Parameter(torch.cat([self.init_value_memory.unsqueeze(0) for _ in range(batch_size)], 0).data)
        self.mem.init_value_memory(value_memory)

        self.q_embed = nn.Embedding(self.n_question + 1, self.key_embedding_dim, padding_idx=0)
        self.qa_embed = nn.Embedding(2 * self.n_question + 1, self.value_embedding_dim, padding_idx=0)

    def init_params(self):
        nn.init.kaiming_normal_(self.predict_linear.weight)
        nn.init.kaiming_normal_(self.read_embed_linear.weight)
        nn.init.constant_(self.read_embed_linear.bias, 0)
        nn.init.constant_(self.predict_linear.bias, 0)

    def init_embeddings(self):

        nn.init.kaiming_normal_(self.q_embed.weight)
        nn.init.kaiming_normal_(self.qa_embed.weight)

    def forward(self, q_data, qa_data, target):

        batch_size = q_data.shape[0]
        seqlen = q_data.shape[1]
        q_embed_data = self.q_embed(q_data)
        qa_embed_data = self.qa_embed(qa_data)

        value_memory = nn.Parameter(torch.cat([self.init_value_memory.unsqueeze(0) for _ in range(batch_size)], 0).data)
        self.mem.init_value_memory(value_memory)

        slice_q_embed_data = torch.chunk(q_embed_data, seqlen, 1)
        slice_qa_embed_data = torch.chunk(qa_embed_data, seqlen, 1)

        value_read_content_l = []
        input_embed_l = []
        for i in range(seqlen):
            # Attention
            q = slice_q_embed_data[i].squeeze(1)
            correlation_weight = self.mem.attention(q)

            # Read Process
            read_content = self.mem.read(correlation_weight)
            value_read_content_l.append(read_content)
            input_embed_l.append(q)
            # Write Process
            qa = slice_qa_embed_data[i].squeeze(1)
            self.mem.write(correlation_weight, qa)

        all_read_value_content = torch.cat([value_read_content_l[i].unsqueeze(1) for i in range(seqlen)], 1)
        input_embed_content = torch.cat([input_embed_l[i].unsqueeze(1) for i in range(seqlen)], 1)

        predict_input = torch.cat([all_read_value_content, input_embed_content], 2)
        read_content_embed = torch.tanh(self.read_embed_linear(predict_input.view(batch_size * seqlen, -1)))

        pred = self.predict_linear(read_content_embed)
        target_1d = target                   # [batch_size * seq_len, 1]
        mask = target_1d.ge(0)               # [batch_size * seq_len, 1]
        pred_1d = pred.view(-1, 1)           # [batch_size * seq_len, 1]

        filtered_pred = torch.masked_select(pred_1d, mask)
        filtered_target = torch.masked_select(target_1d, mask)
        loss = F.binary_cross_entropy_with_logits(filtered_pred, filtered_target)

        return loss, torch.sigmoid(filtered_pred), filtered_target


class DKVMN(KTM):
    def __init__(self, n_question, batch_size, key_embedding_dim, value_embedding_dim,
                 memory_size, key_memory_state_dim, value_memory_state_dim, final_fc_dim, student_num=None):
        super(DKVMN, self).__init__()
        self.batch_size = batch_size
        self.n_question = n_question
        self.model = Net(n_question, batch_size, key_embedding_dim, value_embedding_dim,
                         memory_size, key_memory_state_dim, value_memory_state_dim, final_fc_dim, student_num)

    def train_epoch(self, epoch, model, params, optimizer, q_data, qa_data):
        N = int(math.floor(len(q_data) / params['batch_size']))

        pred_list = []
        target_list = []
        epoch_loss = 0

        model.train()

        for idx in tqdm(range(N), "Epoch %s" % epoch):
            q_one_seq = q_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]
            qa_batch_seq = qa_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]
            target = qa_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]

            target = (target - 1) / params['n_question']
            target = np.floor(target)
            input_q = torch.LongTensor(q_one_seq).to(device)
            input_qa = torch.LongTensor(qa_batch_seq).to(device)
            target = torch.FloatTensor(target).to(device)
            target_to_1d = torch.chunk(target, params['batch_size'], 0)
            target_1d = torch.cat([target_to_1d[i] for i in range(params['batch_size'])], 1)
            target_1d = target_1d.permute(1, 0)

            model.zero_grad()
            loss, filtered_pred, filtered_target = model.forward(input_q, input_qa, target_1d)
            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), params['maxgradnorm'])
            optimizer.step()
            epoch_loss += loss.item()

            right_target = np.asarray(filtered_target.data.tolist())
            right_pred = np.asarray(filtered_pred.data.tolist())
            pred_list.append(right_pred)
            target_list.append(right_target)

        all_pred = np.concatenate(pred_list, axis=0)
        all_target = np.concatenate(target_list, axis=0)
        auc = metrics.roc_auc_score(all_target, all_pred)
        all_pred[all_pred >= 0.5] = 1.0
        all_pred[all_pred < 0.5] = 0.0
        accuracy = metrics.accuracy_score(all_target, all_pred)

        return epoch_loss / N, accuracy, auc

    def train(self, params, train_data, test_data=None):
        q_data, qa_data = train_data

        model = self.model
        model.init_embeddings()
        model.init_params()
        optimizer = torch.optim.Adam(params=model.parameters(), lr=params['lr'], betas=(0.9, 0.9))

        model.to(device)

        all_valid_loss = {}
        all_valid_accuracy = {}
        all_valid_auc = {}
        best_valid_auc = 0

        for idx in range(params['max_iter']):
            train_loss, train_accuracy, train_auc = self.train_epoch(idx, model, params, optimizer, q_data, qa_data)
            print('Epoch %d/%d, loss : %3.5f, auc : %3.5f, accuracy : %3.5f' %
                  (idx + 1, params['max_iter'], train_loss, train_auc, train_accuracy))
            if test_data is not None:
                valid_loss, valid_accuracy, valid_auc = self.eval(params, test_data)
                all_valid_loss[idx + 1] = valid_loss
                all_valid_accuracy[idx + 1] = valid_accuracy
                all_valid_auc[idx + 1] = valid_auc
                # output the epoch with the best validation auc
                if valid_auc > best_valid_auc:
                    print('valid auc improve: %3.4f to %3.4f' % (best_valid_auc, valid_auc))
                    best_valid_auc = valid_auc

    def eval(self, params, data):
        q_data, qa_data = data
        model = self.model
        N = int(math.floor(len(q_data) / params['batch_size']))

        pred_list = []
        target_list = []
        epoch_loss = 0
        model.eval()

        for idx in tqdm(range(N), "Evaluating"):

            q_one_seq = q_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]
            qa_batch_seq = qa_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]
            target = qa_data[idx * params['batch_size']:(idx + 1) * params['batch_size'], :]

            target = (target - 1) / params['n_question']
            target = np.floor(target)

            input_q = torch.LongTensor(q_one_seq).to(device)
            input_qa = torch.LongTensor(qa_batch_seq).to(device)
            target = torch.FloatTensor(target).to(device)

            target_to_1d = torch.chunk(target, params['batch_size'], 0)
            target_1d = torch.cat([target_to_1d[i] for i in range(params['batch_size'])], 1)
            target_1d = target_1d.permute(1, 0)

            loss, filtered_pred, filtered_target = model.forward(input_q, input_qa, target_1d)

            right_target = np.asarray(filtered_target.data.tolist())
            right_pred = np.asarray(filtered_pred.data.tolist())
            pred_list.append(right_pred)
            target_list.append(right_target)
            epoch_loss += loss.item()

        all_pred = np.concatenate(pred_list, axis=0)
        all_target = np.concatenate(target_list, axis=0)

        auc = metrics.roc_auc_score(all_target, all_pred)
        all_pred[all_pred >= 0.5] = 1.0
        all_pred[all_pred < 0.5] = 0.0
        accuracy = metrics.accuracy_score(all_target, all_pred)
        print('valid auc : %3.5f, valid accuracy : %3.5f' % (auc, accuracy))

        return epoch_loss / N, accuracy, auc

    def save(self, filepath):
        torch.save(self.model.state_dict(), filepath)
        logging.info("save parameters to %s" % filepath)

    def load(self, filepath):
        self.model.load_state_dict(torch.load(filepath))
        logging.info("load parameters from %s" % filepath)



# coding: utf-8
# 2021/4/5 @ liujiayu
from EduKTM import EKPT


def test_train(data, tmp_path):
    train_set, q_m, stu_num, prob_num, know_num, time_window_num = data

    cdm = EKPT(q_m, stu_num, prob_num, know_num, time_window_num=time_window_num)

    cdm.train(train_set, epoch=30, lr=0.001, lr_b=0.0001, epsilon=1e-1, init_method='mean')
    rmse, mae = cdm.eval([{'user_id': 0, 'item_id': 0, 'score': 1.0}])
    filepath = tmp_path / "ekpt.params"
    cdm.save(filepath)
    cdm.load(filepath)



# coding: utf-8
# 2021/5/24 @ tongshiwei
__all__ = ["pick", "tensor2list", "length2mask", "get_sequence_mask", "sequence_mask"]

import torch
from torch import Tensor


def pick(tensor, index, axis=-1):
    return torch.gather(tensor, axis, index.unsqueeze(axis)).squeeze(axis)


def tensor2list(tensor: Tensor):
    return tensor.cpu().tolist()


def length2mask(length, max_len, valid_mask_val, invalid_mask_val):
    mask = []

    if isinstance(valid_mask_val, Tensor):
        valid_mask_val = tensor2list(valid_mask_val)
    if isinstance(invalid_mask_val, Tensor):
        invalid_mask_val = tensor2list(invalid_mask_val)
    if isinstance(length, Tensor):
        length = tensor2list(length)

    for _len in length:
        mask.append([valid_mask_val] * _len + [invalid_mask_val] * (max_len - _len))

    return torch.tensor(mask)


def get_sequence_mask(shape, sequence_length, axis=1):
    assert axis <= len(shape)
    mask_shape = shape[axis + 1:]

    valid_mask_val = torch.ones(mask_shape)
    invalid_mask_val = torch.zeros(mask_shape)

    max_len = shape[axis]

    return length2mask(sequence_length, max_len, valid_mask_val, invalid_mask_val)


def sequence_mask(tensor: Tensor, sequence_length, axis=1):
    mask = get_sequence_mask(tensor.shape, sequence_length, axis).to(tensor.device)
    return tensor * mask



# coding: utf-8
# 2021/3/18 @ tongshiwei

from .meta import KTM
from .KPT import KPT
from .EKPT import EKPT
from .DKT import DKT
from .DKTPlus import DKTPlus
from .AKT import AKT
from .LPKT import LPKT
from .GKT import GKT
from .DKVMN import DKVMN



# coding: utf-8
# 2022/3/1 @ fannazya
__all__ = ["GKTNet"]

import json
import networkx as nx
import torch
from torch import nn
import torch.nn.functional as F
from EduKTM.utils import GRUCell, begin_states, get_states, expand_tensor, \
    format_sequence, mask_sequence_variable_length


class GKTNet(nn.Module):
    def __init__(self, ku_num, graph, hidden_num=None, latent_dim=None, dropout=0.0):
        super(GKTNet, self).__init__()
        self.ku_num = int(ku_num)
        self.hidden_num = self.ku_num if hidden_num is None else int(hidden_num)
        self.latent_dim = self.ku_num if latent_dim is None else int(latent_dim)
        self.neighbor_dim = self.hidden_num + self.latent_dim
        self.graph = nx.DiGraph()
        self.graph.add_nodes_from(list(range(ku_num)))
        try:
            with open(graph) as f:
                self.graph.add_weighted_edges_from(json.load(f))
        except ValueError:
            with open(graph) as f:
                self.graph.add_weighted_edges_from([e + [1.0] for e in json.load(f)])

        self.rnn = GRUCell(self.hidden_num)
        self.response_embedding = nn.Embedding(2 * self.ku_num, self.latent_dim)
        self.concept_embedding = nn.Embedding(self.ku_num, self.latent_dim)
        self.f_self = nn.Linear(self.neighbor_dim, self.hidden_num)
        self.n_out = nn.Linear(2 * self.neighbor_dim, self.hidden_num)
        self.n_in = nn.Linear(2 * self.neighbor_dim, self.hidden_num)
        self.dropout = nn.Dropout(dropout)
        self.out = nn.Linear(self.hidden_num, 1)

    def in_weight(self, x, ordinal=True, with_weight=True):
        if isinstance(x, torch.Tensor):
            x = x.numpy().tolist()
        if isinstance(x, list):
            return [self.in_weight(_x) for _x in x]
        elif isinstance(x, (int, float)):
            if not ordinal:
                return list(self.graph.predecessors(int(x)))
            else:
                _ret = [0] * self.ku_num
                for i in self.graph.predecessors(int(x)):
                    if with_weight:
                        _ret[i] = self.graph[i][x]['weight']
                    else:
                        _ret[i] = 1
                return _ret
        else:
            raise TypeError("cannot handle %s" % type(x))

    def out_weight(self, x, ordinal=True, with_weight=True):
        if isinstance(x, torch.Tensor):
            x = x.numpy().tolist()
        if isinstance(x, list):
            return [self.out_weight(_x) for _x in x]
        elif isinstance(x, (int, float)):
            if not ordinal:
                return list(self.graph.successors(int(x)))
            else:
                _ret = [0] * self.ku_num
                for i in self.graph.successors(int(x)):
                    if with_weight:
                        _ret[i] = self.graph[x][i]['weight']
                    else:
                        _ret[i] = 1
                return _ret
        else:
            raise TypeError("cannot handle %s" % type(x))

    def neighbors(self, x, ordinal=True, with_weight=False):
        if isinstance(x, torch.Tensor):
            x = x.numpy().tolist()
        if isinstance(x, list):
            return [self.neighbors(_x) for _x in x]
        elif isinstance(x, (int, float)):
            if not ordinal:
                return list(self.graph.neighbors(int(x)))
            else:
                _ret = [0] * self.ku_num
                for i in self.graph.neighbors(int(x)):
                    if with_weight:
                        _ret[i] = self.graph[i][x]['weight']
                    else:
                        _ret[i] = 1
                return _ret
        else:
            raise TypeError("cannot handle %s" % type(x))

    def forward(self, questions, answers, valid_length=None, compressed_out=True, layout="NTC"):
        length = questions.shape[1]
        inputs, axis, batch_size = format_sequence(length, questions, layout, False)
        answers, _, _ = format_sequence(length, answers, layout, False)

        states = begin_states([(batch_size, self.ku_num, self.hidden_num)])[0]
        outputs = []
        all_states = []
        for i in range(length):
            # neighbors - aggregate
            inputs_i = inputs[i].reshape([batch_size, ])
            answer_i = answers[i].reshape([batch_size, ])

            _neighbors = self.neighbors(inputs_i)
            neighbors_mask = expand_tensor(torch.Tensor(_neighbors), -1, self.hidden_num)
            _neighbors_mask = expand_tensor(torch.Tensor(_neighbors), -1, self.hidden_num + self.latent_dim)

            # get concept embedding
            concept_embeddings = self.concept_embedding.weight.data
            concept_embeddings = expand_tensor(concept_embeddings, 0, batch_size)

            agg_states = torch.cat((concept_embeddings, states), dim=-1)

            # aggregate
            _neighbors_states = _neighbors_mask * agg_states

            # self - aggregate
            _concept_embedding = get_states(inputs_i, states)
            _self_hidden_states = torch.cat((_concept_embedding, self.response_embedding(answer_i)), dim=-1)

            _self_mask = F.one_hot(inputs_i, self.ku_num)  # p
            _self_mask = expand_tensor(_self_mask, -1, self.hidden_num)

            self_hidden_states = expand_tensor(_self_hidden_states, 1, self.ku_num)

            # aggregate
            _hidden_states = torch.cat((_neighbors_states, self_hidden_states), dim=-1)

            _in_state = self.n_in(_hidden_states)
            _out_state = self.n_out(_hidden_states)
            in_weight = expand_tensor(torch.Tensor(self.in_weight(inputs_i)), -1, self.hidden_num)
            out_weight = expand_tensor(torch.Tensor(self.out_weight(inputs_i)), -1, self.hidden_num)

            next_neighbors_states = in_weight * _in_state + out_weight * _out_state

            # self - update
            next_self_states = self.f_self(_self_hidden_states)
            next_self_states = expand_tensor(next_self_states, 1, self.ku_num)
            next_self_states = _self_mask * next_self_states

            next_states = neighbors_mask * next_neighbors_states + next_self_states

            next_states, _ = self.rnn(next_states, [states])
            next_states = (_self_mask + neighbors_mask) * next_states + (1 - _self_mask - neighbors_mask) * states

            states = self.dropout(next_states)
            output = torch.sigmoid(self.out(states).squeeze(axis=-1))  # p
            outputs.append(output)
            if valid_length is not None and not compressed_out:
                all_states.append([states])

        if valid_length is not None:
            if compressed_out:
                states = None
            outputs = mask_sequence_variable_length(torch, outputs, length, valid_length, axis, merge=True)

        return outputs, states



# coding: utf-8
# 2021/4/5 @ liujiayu

import logging
import numpy as np
import pickle
from tqdm import tqdm
from collections import namedtuple
from collections import defaultdict
from EduKTM import KTM

hyper_para = namedtuple("hyperparameters", ["r", "D", "deltaT", "S", "lambda_U_1", "lambda_U", "lambda_P", "lambda_S"])
default_hyper = hyper_para(6, 2, 1, 5, 0.01, 2, 2, 0.01)


def init_parameters(stu_num, prob_num, know_num, time_window_num):
    u_latent = np.random.normal(0.5, 0.01, size=(time_window_num, stu_num, know_num))
    i_latent = 0.1 * np.random.uniform(0, 1, size=(prob_num, know_num))  # problems' latent vector(V)
    alpha = np.random.uniform(0, 1, size=stu_num)
    B = 0.01 * np.random.normal(0, 1, size=prob_num)
    return u_latent, i_latent, alpha, B


def stu_curve(u_latent, alpha, r, D, deltaT, S, time_freq):  # learning and forgetting curve
    freq_norm = D * time_freq / (time_freq + r)
    learn_factor = u_latent * freq_norm
    forget_factor = u_latent * np.exp(-deltaT / S)
    pred_u = learn_factor * np.expand_dims(alpha, axis=1) + forget_factor * np.expand_dims(1 - alpha, axis=1)
    return pred_u, freq_norm


class EKPT(KTM):
    """
    EKPT model, training (MAP) and testing methods
    Parameters
    ----------
    q_m: array
        Q matrix, shape = (prob_num, know_num)
    stu_num: int
        number of students
    prob_num: int
        number of problems
    know_num: int
        number of knowledge
    time_window_num: int
        number of time windows
    args: namedtuple
        all hyper-parameters
    ----------
    """

    def __init__(self, q_m, stu_num, prob_num, know_num, time_window_num, args=default_hyper):
        super(EKPT, self).__init__()
        self.args = args
        self.q_m = q_m
        self.stu_num, self.prob_num, self.know_num = stu_num, prob_num, know_num
        self.time_window_num = time_window_num
        self.u_latent, self.i_latent, self.alpha, self.B = init_parameters(stu_num, prob_num, know_num, time_window_num)
        # partial order of knowledge in each problem
        self.par_mat = np.zeros(shape=(prob_num, know_num, know_num))
        for i in range(prob_num):
            for o1 in range(know_num):
                if self.q_m[i][o1] == 0:
                    continue
                for o2 in range(know_num):
                    if self.q_m[i][o2] == 0:
                        self.par_mat[i][o1][o2] = 1

        # exercise relation
        self.exer_neigh = (np.dot(self.q_m, self.q_m.transpose()) > 0).astype(int)

    def train(self, train_data, epoch, lr=0.001, lr_b=0.0001, epsilon=1e-3, init_method='mean') -> ...:
        # train_data(list): response data, length = time_window_num, e.g.[[{'user_id':, 'item_id':, 'score':},...],...]
        assert self.time_window_num == len(train_data), 'number of time windows conflicts'
        u_latent, i_latent = np.copy(self.u_latent), np.copy(self.i_latent)
        alpha, B = np.copy(self.alpha), np.copy(self.B)
        # mean score of each student in train_data
        sum_score = np.zeros(shape=self.stu_num)
        sum_count = np.zeros(shape=self.stu_num)

        # knowledge frequency in each time window
        time_freq = np.zeros(shape=(self.time_window_num, self.stu_num, self.know_num))
        for t in range(self.time_window_num):
            for record in train_data[t]:
                user, item, rating = record['user_id'], record['item_id'], record['score']
                time_freq[t][user][np.where(self.q_m[item] == 1)[0]] += 1
                sum_score[user] += rating
                sum_count[user] += 1

        # initialize student latent with mean score
        if init_method == 'mean':
            u_latent = np.random.normal(20 * np.expand_dims(sum_score / (sum_count + 1e-9), axis=1) / self.know_num,
                                        0.01, size=(self.time_window_num, self.stu_num, self.know_num))

        for iteration in range(epoch):
            u_latent_tmp, i_latent_tmp = np.copy(u_latent), np.copy(i_latent)
            alpha_tmp, B_tmp = np.copy(alpha), np.copy(B)
            i_gradient = np.zeros(shape=(self.prob_num, self.know_num))
            b_gradient = np.zeros(shape=self.prob_num)
            alpha_gradient = np.zeros(shape=self.stu_num)
            for t in range(self.time_window_num):
                u_gradient_t = np.zeros(shape=(self.stu_num, self.know_num))
                record_num_t = len(train_data[t])
                users = [record['user_id'] for record in train_data[t]]
                items = [record['item_id'] for record in train_data[t]]
                ratings = [record['score'] for record in train_data[t]]

                pred_R = [np.dot(u_latent[t][users[i]], i_latent[items[i]]) - B[items[i]] for i in range(record_num_t)]
                pred_u, freq_norm = stu_curve(u_latent, alpha, self.args.r, self.args.D, self.args.deltaT, self.args.S,
                                              time_freq)  # both shape are (time_window_num, stu_num, know_num)
                for i in range(record_num_t):
                    user, item, rating = users[i], items[i], ratings[i]
                    R_diff = pred_R[i] - rating
                    b_gradient[item] -= R_diff
                    u_gradient_t[user] += R_diff * i_latent[item]
                    i_gradient[item] += R_diff * u_latent[t][user] + self.args.lambda_S * i_latent[item]
                    i_gradient[item] -= self.args.lambda_S * np.sum(
                        np.expand_dims(self.exer_neigh[item], axis=1) * i_latent, axis=0) / sum(self.exer_neigh[item])
                    if t == 0:
                        u_gradient_t[user] += self.args.lambda_U_1 * u_latent[0][user]
                    else:
                        u_gradient_t[user] += self.args.lambda_U * (u_latent[t][user] - pred_u[t - 1][user])
                        alpha_gradient[user] += np.dot(pred_u[t - 1][user] - u_latent[t][user], u_latent[t][user] * (
                            freq_norm[t - 1][user] - np.exp(-self.args.deltaT / self.args.S)))
                    if t < self.time_window_num - 1:
                        u_gradient_t[user] += self.args.lambda_U * (pred_u[t][user] - u_latent[t + 1][user]) * (
                            alpha[user] * freq_norm[t][user] + (1 - alpha[user]) * np.exp(
                                - self.args.deltaT / self.args.S))
                    o1, o2 = np.where(self.par_mat[item] == 1)
                    for j in range(len(o1)):
                        i_gradient[item][o1[j]] -= self.args.lambda_P * 0.5 * (1 - np.tanh(
                            0.5 * (i_latent[item][o1[j]] - i_latent[item][o2[j]])))
                        i_gradient[item][o2[j]] += self.args.lambda_P * 0.5 * (1 - np.tanh(
                            0.5 * (i_latent[item][o1[j]] - i_latent[item][o2[j]])))
                u_latent[t] -= lr * u_gradient_t
            i_latent -= lr * i_gradient
            B -= lr_b * b_gradient
            alpha = np.clip(alpha - lr * alpha_gradient, 0, 1)
            change = max(np.max(np.abs(u_latent - u_latent_tmp)), np.max(np.abs(i_latent - i_latent_tmp)),
                         np.max(np.abs(alpha - alpha_tmp)), np.max(np.abs(B - B_tmp)))
            if iteration > 20 and change < epsilon:
                break
        self.u_latent, self.i_latent, self.alpha, self.B = u_latent, i_latent, alpha, B

    def eval(self, test_data) -> tuple:
        test_rmse, test_mae = [], []
        for i in tqdm(test_data, "evaluating"):
            stu, test_id, true_score = i['user_id'], i['item_id'], i['score']
            predict_rating = np.clip(np.dot(self.u_latent[-1][stu], self.i_latent[test_id]) - self.B[test_id], 0, 1)
            test_rmse.append((predict_rating - true_score) ** 2)
            test_mae.append(abs(predict_rating - true_score))
        return np.sqrt(np.average(test_rmse)), np.average(test_mae)

    def save(self, filepath):
        with open(filepath, 'wb') as file:
            pickle.dump({"U": self.u_latent, "V": self.i_latent, "alpha": self.alpha, "B": self.B}, file)
            logging.info("save parameters to %s" % filepath)

    def load(self, filepath):
        with open(filepath, 'rb') as file:
            self.u_latent, self.i_latent, self.alpha, self.B = pickle.load(file).values()
            logging.info("load parameters from %s" % filepath)



# coding: utf-8
# 2021/3/18 @ tongshiwei


def etl(*args, **kwargs) -> ...:  # pragma: no cover
    """
    extract - transform - load
    """
    pass


def train(*args, **kwargs) -> ...:  # pragma: no cover
    pass


def evaluate(*args, **kwargs) -> ...:  # pragma: no cover
    pass


class KTM(object):
    def __init__(self, *args, **kwargs) -> ...:
        pass

    def train(self, *args, **kwargs) -> ...:
        raise NotImplementedError

    def eval(self, *args, **kwargs) -> ...:
        raise NotImplementedError

    def save(self, *args, **kwargs) -> ...:
        raise NotImplementedError

    def load(self, *args, **kwargs) -> ...:
        raise NotImplementedError



# coding: utf-8
# 2021/8/17 @ sone

import torch
from torch import nn

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class LPKTNet(nn.Module):
    def __init__(self, n_at, n_it, n_exercise, n_question, d_a, d_e, d_k, q_matrix, dropout=0.2):
        super(LPKTNet, self).__init__()
        self.d_k = d_k
        self.d_a = d_a
        self.d_e = d_e
        self.q_matrix = q_matrix
        self.n_question = n_question

        self.at_embed = nn.Embedding(n_at + 10, d_k)
        torch.nn.init.xavier_uniform_(self.at_embed.weight)
        self.it_embed = nn.Embedding(n_it + 10, d_k)
        torch.nn.init.xavier_uniform_(self.it_embed.weight)
        self.e_embed = nn.Embedding(n_exercise + 10, d_k)
        torch.nn.init.xavier_uniform_(self.e_embed.weight)

        self.linear_1 = nn.Linear(d_a + d_e + d_k, d_k)
        torch.nn.init.xavier_uniform_(self.linear_1.weight)
        self.linear_2 = nn.Linear(4 * d_k, d_k)
        torch.nn.init.xavier_uniform_(self.linear_2.weight)
        self.linear_3 = nn.Linear(4 * d_k, d_k)
        torch.nn.init.xavier_uniform_(self.linear_3.weight)
        self.linear_4 = nn.Linear(3 * d_k, d_k)
        torch.nn.init.xavier_uniform_(self.linear_4.weight)
        self.linear_5 = nn.Linear(d_e + d_k, d_k)
        torch.nn.init.xavier_uniform_(self.linear_5.weight)

        self.tanh = nn.Tanh()
        self.sig = nn.Sigmoid()
        self.dropout = nn.Dropout(dropout)

    def forward(self, e_data, at_data, a_data, it_data):
        batch_size, seq_len = e_data.size(0), e_data.size(1)
        e_embed_data = self.e_embed(e_data)
        at_embed_data = self.at_embed(at_data)
        it_embed_data = self.it_embed(it_data)
        a_data = a_data.view(-1, 1).repeat(1, self.d_a).view(batch_size, -1, self.d_a)
        h_pre = nn.init.xavier_uniform_(torch.zeros(self.n_question + 1, self.d_k)).repeat(batch_size, 1, 1).to(device)
        h_tilde_pre = None
        all_learning = self.linear_1(torch.cat((e_embed_data, at_embed_data, a_data), 2))
        learning_pre = torch.zeros(batch_size, self.d_k).to(device)

        pred = torch.zeros(batch_size, seq_len).to(device)

        for t in range(0, seq_len - 1):
            e = e_data[:, t]
            # q_e: (bs, 1, n_skill)
            q_e = self.q_matrix[e].view(batch_size, 1, -1)
            it = it_embed_data[:, t]

            # Learning Module
            if h_tilde_pre is None:
                h_tilde_pre = q_e.bmm(h_pre).view(batch_size, self.d_k)
            learning = all_learning[:, t]
            learning_gain = self.linear_2(torch.cat((learning_pre, it, learning, h_tilde_pre), 1))
            learning_gain = self.tanh(learning_gain)
            gamma_l = self.linear_3(torch.cat((learning_pre, it, learning, h_tilde_pre), 1))
            gamma_l = self.sig(gamma_l)
            LG = gamma_l * ((learning_gain + 1) / 2)
            LG_tilde = self.dropout(q_e.transpose(1, 2).bmm(LG.view(batch_size, 1, -1)))

            # Forgetting Module
            # h_pre: (bs, n_skill, d_k)
            # LG: (bs, d_k)
            # it: (bs, d_k)
            n_skill = LG_tilde.size(1)
            gamma_f = self.sig(self.linear_4(torch.cat((
                h_pre,
                LG.repeat(1, n_skill).view(batch_size, -1, self.d_k),
                it.repeat(1, n_skill).view(batch_size, -1, self.d_k)
            ), 2)))
            h = LG_tilde + gamma_f * h_pre

            # Predicting Module
            h_tilde = self.q_matrix[e_data[:, t + 1]].view(batch_size, 1, -1).bmm(h).view(batch_size, self.d_k)
            y = self.sig(self.linear_5(torch.cat((e_embed_data[:, t + 1], h_tilde), 1))).sum(1) / self.d_k
            pred[:, t + 1] = y

            # prepare for next prediction
            learning_pre = learning
            h_pre = h
            h_tilde_pre = h_tilde

        return pred



# coding: utf-8
# 2021/8/21 @ zengxiaonan

from EduKTM import LPKT


def test_train(data, conf, tmp_path):
    n_at, n_it, n_question, n_exercise, q_matrix, batch_size = conf
    d_a = 16
    d_e = 32
    d_k = 32
    dropout = 0.2

    lpkt = LPKT(n_at, n_it, n_exercise, n_question, d_a, d_e, d_k, q_matrix, batch_size, dropout)
    lpkt.train(data, test_data=data, epoch=2)
    filepath = tmp_path / "lpkt.params"
    lpkt.save(filepath)
    lpkt.load(filepath)



# coding: utf-8
# 2022/2/25 @ fannazya
# Code reused from https://github.com/tswsxk/Baize
__all__ = ["GRUCell", "begin_states", "get_states", "expand_tensor",
           "mask_sequence_variable_length", "format_sequence"]

import torch
from torch import nn


def as_list(obj) -> list:
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    else:
        return [obj]


def begin_states(shapes, func=torch.zeros):
    states = []
    for i, shape in enumerate(as_list(shapes)):
        state = func(shape)
        states.append(state)
    return states


def get_states(indexes, states):
    if isinstance(indexes, torch.Tensor):
        indexes = indexes.numpy().tolist()
    if isinstance(indexes, list):
        return torch.stack([*[get_states(index, state) for (index, state) in zip(indexes, states)]])
    elif isinstance(indexes, (int, float)):
        return states[int(indexes)]
    else:
        raise TypeError("cannot handle %s" % type(indexes))


def sequence_mask(X, valid_len, value):
    maxlen = X.size(1)
    mask = torch.arange(maxlen, dtype=torch.float32,
                        device=X.device)[None, :] < valid_len[:, None]
    X[~mask] = value

    return X


def mask_sequence_variable_length(F, data, length, valid_length, time_axis,
                                  merge):
    assert valid_length is not None
    if not isinstance(data, torch.Tensor):
        data = F.stack([*data], dim=1)
    outputs = sequence_mask(
        data, valid_len=valid_length,
        value=0
    )
    if not merge:
        outputs = as_list(
            F.split(
                outputs, split_size_or_sections=length, axis=time_axis
            )
        )
    return outputs


def format_sequence(length, inputs, layout, merge, in_layout=None):

    assert inputs is not None, \
        "unroll(inputs=None) has been deprecated. " \
        "Please create input variables outside unroll."

    axis = layout.find('T')
    batch_axis = layout.find('N')
    batch_size = 0
    in_axis = in_layout.find('T') if in_layout is not None else axis
    if isinstance(inputs, torch.Tensor):
        batch_size = inputs.shape[batch_axis]
        if merge is False:
            assert length is None or length == inputs.shape[in_axis]
            split_size = [1] * inputs.shape[in_axis]
            inputs = as_list(torch.split(inputs, split_size_or_sections=split_size, dim=in_axis))

    return inputs, axis, batch_size


def expand_tensor(tensor, expand_axis, expand_num):
    assert len(tensor.shape) == 2

    _tensor = torch.unsqueeze(tensor, expand_axis)

    shape = list(_tensor.shape)
    shape[expand_axis] = expand_num
    _tensor = torch.broadcast_to(_tensor, tuple(shape))

    return _tensor


class GRUCell(nn.Module):
    def __init__(self, hidden_num):
        super(GRUCell, self).__init__()
        self.i2h = nn.Linear(hidden_num, 3 * hidden_num)
        self.h2h = nn.Linear(hidden_num, 3 * hidden_num)
        self.reset_act = nn.Sigmoid()
        self.update_act = nn.Sigmoid()
        self.act = nn.Tanh()

    def forward(self, inputs, states):
        prev_state_h = states[0]

        i2h = self.i2h(inputs)
        h2h = self.h2h(prev_state_h)
        i2h_r, i2h_z, i2h = torch.chunk(i2h, 3, dim=-1)
        h2h_r, h2h_z, h2h = torch.chunk(h2h, 3, dim=-1)

        reset_gate = self.reset_act(i2h_r + h2h_r)
        update_gate = self.update_act(i2h_z + h2h_z)
        next_h_tmp = self.act(i2h + reset_gate * h2h)
        ones = torch.ones_like(update_gate)
        next_h = (ones - update_gate) * next_h_tmp + update_gate * prev_state_h

        return next_h, [next_h]



# coding: utf-8
# 2021/4/5 @ liujiayu

import random
import numpy as np
import pytest


@pytest.fixture(scope="package")
def conf():
    user_num = 5
    item_num = 2
    know_num = 3
    time_window_num = 2
    return user_num, item_num, know_num, time_window_num


@pytest.fixture(scope="package")
def data(conf):
    user_num, item_num, know_num, time_window_num = conf
    q_m = np.zeros(shape=(item_num, know_num))
    for i in range(item_num):
        q_m[i, random.randint(0, know_num - 1)] = 1

    train_set = []
    for t in range(time_window_num):
        train_t = []
        for u in range(user_num):
            item_t = random.randint(0, item_num - 1)
            rating_t = random.randint(0, 1)
            train_t.append({'user_id': u, 'item_id': item_t, 'score': rating_t})
        train_set.append(train_t)

    return train_set, q_m, user_num, item_num, know_num, time_window_num



# coding: utf-8
# 2022/3/18 @ ouyangjie


import pytest
from EduKTM import DKVMN


@pytest.mark.parametrize("lr", [0, 0.1])
@pytest.mark.parametrize("maxgradnorm", [-5, 5])
def test_train(data, conf, tmp_path, lr, maxgradnorm):
    n_question, batch_size = conf

    params = {
        'max_iter': 2,
        'lr': lr,
        'final_fc_dim': 5,
        'key_embedding_dim': 5,
        'batch_size': batch_size,
        'value_embedding_dim': 10,
        'memory_size': 5,
        'n_question': n_question,
        'seqlen': 10,
        'maxgradnorm': maxgradnorm
    }

    dkvmn = DKVMN(n_question=params['n_question'],
                  batch_size=params['batch_size'],
                  key_embedding_dim=params['key_embedding_dim'],
                  value_embedding_dim=params['value_embedding_dim'],
                  memory_size=params['memory_size'],
                  key_memory_state_dim=params['key_embedding_dim'],
                  value_memory_state_dim=params['value_embedding_dim'],
                  final_fc_dim=params['final_fc_dim'])

    dkvmn.train(params, data, test_data=data)
    filepath = tmp_path / "dkvmn.params"
    dkvmn.save(filepath)
    dkvmn.load(filepath)
    dkvmn.eval(params, data)



# coding: utf-8
# 2022/3/18 @ ouyangjie


import random
import pytest
import numpy as np


@pytest.fixture(scope="package")
def conf():
    n_question = 10
    batch_size = 4
    return n_question, batch_size


@pytest.fixture(scope="package")
def data(conf):
    n_question, batch_size = conf
    seqlen = 10

    q = [
        [random.randint(1, n_question) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    a = [
        [random.randint(0, 1) for _ in range(seqlen)]
        for _ in range(batch_size)
    ]
    qa = [
        [a[i][j] * n_question + q_id for j, q_id in enumerate(q_seq)]
        for i, q_seq in enumerate(q)
    ]
    data = (np.array(q), np.array(qa))

    return data



# coding: utf-8
# 2021/8/17 @ sone

import math
import logging
import torch
import torch.nn as nn
import numpy as np
from sklearn import metrics
import tqdm

from EduKTM import KTM
from .LPKTNet import LPKTNet

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def binary_entropy(target, pred):
    loss = target * np.log(np.maximum(1e-10, pred)) + (1.0 - target) * np.log(np.maximum(1e-10, 1.0 - pred))
    return np.average(loss) * -1.0


def compute_auc(all_target, all_pred):
    return metrics.roc_auc_score(all_target, all_pred)


def compute_accuracy(all_target, all_pred):
    all_pred[all_pred > 0.5] = 1.0
    all_pred[all_pred <= 0.5] = 0.0
    return metrics.accuracy_score(all_target, all_pred)


def train_one_epoch(net, optimizer, criterion, batch_size, a_data, e_data, it_data, at_data):
    net.train()
    n = int(math.ceil(len(e_data) / batch_size))
    shuffled_ind = np.arange(e_data.shape[0])
    np.random.shuffle(shuffled_ind)
    e_data = e_data[shuffled_ind]
    at_data = at_data[shuffled_ind]
    a_data = a_data[shuffled_ind]
    it_data = it_data[shuffled_ind]

    pred_list = []
    target_list = []

    for idx in tqdm.tqdm(range(n), 'Training'):
        optimizer.zero_grad()

        e_one_seq = e_data[idx * batch_size: (idx + 1) * batch_size, :]
        at_one_seq = at_data[idx * batch_size: (idx + 1) * batch_size, :]
        a_one_seq = a_data[idx * batch_size: (idx + 1) * batch_size, :]
        it_one_seq = it_data[idx * batch_size: (idx + 1) * batch_size, :]

        input_e = torch.from_numpy(e_one_seq).long().to(device)
        input_at = torch.from_numpy(at_one_seq).long().to(device)
        input_it = torch.from_numpy(it_one_seq).long().to(device)
        target = torch.from_numpy(a_one_seq).float().to(device)

        pred = net(input_e, input_at, target, input_it)

        mask = input_e[:, 1:] > 0
        masked_pred = pred[:, 1:][mask]
        masked_truth = target[:, 1:][mask]

        loss = criterion(masked_pred, masked_truth).sum()

        loss.backward()
        optimizer.step()

        masked_pred = masked_pred.detach().cpu().numpy()
        masked_truth = masked_truth.detach().cpu().numpy()

        pred_list.append(masked_pred)
        target_list.append(masked_truth)

    all_pred = np.concatenate(pred_list, axis=0)
    all_target = np.concatenate(target_list, axis=0)

    loss = binary_entropy(all_target, all_pred)
    auc = compute_auc(all_target, all_pred)
    accuracy = compute_accuracy(all_target, all_pred)

    return loss, auc, accuracy


def test_one_epoch(net, batch_size, a_data, e_data, it_data, at_data):
    net.eval()
    n = int(math.ceil(len(e_data) / batch_size))

    pred_list = []
    target_list = []

    for idx in tqdm.tqdm(range(n), 'Testing'):
        e_one_seq = e_data[idx * batch_size: (idx + 1) * batch_size, :]
        at_one_seq = at_data[idx * batch_size: (idx + 1) * batch_size, :]
        a_one_seq = a_data[idx * batch_size: (idx + 1) * batch_size, :]
        it_one_seq = it_data[idx * batch_size: (idx + 1) * batch_size, :]

        input_e = torch.from_numpy(e_one_seq).long().to(device)
        input_at = torch.from_numpy(at_one_seq).long().to(device)
        input_it = torch.from_numpy(it_one_seq).long().to(device)
        target = torch.from_numpy(a_one_seq).float().to(device)

        with torch.no_grad():
            pred = net(input_e, input_at, target, input_it)

            mask = input_e[:, 1:] > 0
            masked_pred = pred[:, 1:][mask].detach().cpu().numpy()
            masked_truth = target[:, 1:][mask].detach().cpu().numpy()

            pred_list.append(masked_pred)
            target_list.append(masked_truth)

    all_pred = np.concatenate(pred_list, axis=0)
    all_target = np.concatenate(target_list, axis=0)

    loss = binary_entropy(all_target, all_pred)
    auc = compute_auc(all_target, all_pred)
    accuracy = compute_accuracy(all_target, all_pred)

    return loss, auc, accuracy


class LPKT(KTM):
    def __init__(self, n_at, n_it, n_exercise, n_question, d_a, d_e, d_k, q_matrix, batch_size, dropout=0.2):
        super(LPKT, self).__init__()
        q_matrix = torch.from_numpy(q_matrix).float().to(device)
        self.lpkt_net = LPKTNet(n_at, n_it, n_exercise, n_question, d_a, d_e, d_k, q_matrix, dropout).to(device)
        self.batch_size = batch_size

    def train(self, train_data, test_data=None, *, epoch: int, lr=0.002, lr_decay_step=15, lr_decay_rate=0.5) -> ...:
        optimizer = torch.optim.Adam(self.lpkt_net.parameters(), lr=lr, eps=1e-8, betas=(0.1, 0.999), weight_decay=1e-6)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, lr_decay_step, gamma=lr_decay_rate)
        criterion = nn.BCELoss(reduction='none')
        best_train_auc, best_test_auc = .0, .0

        for idx in range(epoch):
            train_loss, train_auc, train_accuracy = train_one_epoch(self.lpkt_net, optimizer, criterion,
                                                                    self.batch_size, *train_data)
            print("[Epoch %d] LogisticLoss: %.6f" % (idx, train_loss))
            if train_auc > best_train_auc:
                best_train_auc = train_auc

            scheduler.step()

            if test_data is not None:
                test_loss, test_auc, test_accuracy = self.eval(test_data)
                print("[Epoch %d] auc: %.6f, accuracy: %.6f" % (idx, test_auc, test_accuracy))
                if test_auc > best_test_auc:
                    best_test_auc = test_auc

        return best_train_auc, best_test_auc

    def eval(self, test_data) -> ...:
        self.lpkt_net.eval()
        return test_one_epoch(self.lpkt_net, self.batch_size, *test_data)

    def save(self, filepath) -> ...:
        torch.save(self.lpkt_net.state_dict(), filepath)
        logging.info("save parameters to %s" % filepath)

    def load(self, filepath) -> ...:
        self.lpkt_net.load_state_dict(torch.load(filepath))
        logging.info("load parameters from %s" % filepath)



# coding: utf-8
# 2021/5/25 @ tongshiwei
# These codes are modified from gluonnlp

__all__ = ["PadSequence"]


class PadSequence:
    """Pad the sequence.

    Pad the sequence to the given `length` by inserting `pad_val`. If `clip` is set,
    sequence that has length larger than `length` will be clipped.

    Parameters
    ----------
    length : int
        The maximum length to pad/clip the sequence
    pad_val : number
        The pad value. Default 0
    clip : bool
    """

    def __init__(self, length, pad_val=0, clip=True):
        self._length = length
        self._pad_val = pad_val
        self._clip = clip

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample : list of number or mx.nd.NDArray or np.ndarray

        Returns
        -------
        ret : list of number or mx.nd.NDArray or np.ndarray
        """
        sample_length = len(sample)
        if sample_length >= self._length:
            if self._clip and sample_length > self._length:
                return sample[:self._length]
            else:
                return sample
        else:
            return sample + [
                self._pad_val for _ in range(self._length - sample_length)
            ]



# coding: utf-8
# 2021/5/24 @ tongshiwei


__all__ = ["as_list"]


def as_list(obj) -> list:
    r"""A utility function that converts the argument to a list
    if it is not already.

    Parameters
    ----------
    obj : object
        argument to be converted to a list

    Returns
    -------
    list_obj: list
        If `obj` is a list or tuple, return it. Otherwise,
        return `[obj]` as a single-element list.

    Examples
    --------
    >>> as_list(1)
    [1]
    >>> as_list([1])
    [1]
    >>> as_list((1, 2))
    [1, 2]
    """
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    else:
        return [obj]



# coding: utf-8
# 2022/2/25 @ fannazya

import logging
import numpy as np
import torch
from tqdm import tqdm
from EduKTM import KTM
from .GKTNet import GKTNet
from EduKTM.utils import SLMLoss, tensor2list, pick
from sklearn.metrics import roc_auc_score, accuracy_score


class GKT(KTM):
    def __init__(self, ku_num, graph, hidden_num, net_params: dict = None, loss_params=None):
        super(GKT, self).__init__()
        self.gkt_model = GKTNet(
            ku_num,
            graph,
            hidden_num,
            **(net_params if net_params is not None else {})
        )
        # self.gkt_model = GKTNet(ku_num, graph, hidden_num)
        self.loss_params = loss_params if loss_params is not None else {}

    def train(self, train_data, test_data=None, *, epoch: int, device="cpu", lr=0.001) -> ...:
        loss_function = SLMLoss(**self.loss_params)
        trainer = torch.optim.Adam(self.gkt_model.parameters(), lr)

        for e in range(epoch):
            losses = []
            for (question, data, data_mask, label, pick_index, label_mask) in tqdm(train_data, "Epoch %s" % e):
                # convert to device
                question: torch.Tensor = question.to(device)
                data: torch.Tensor = data.to(device)
                data_mask: torch.Tensor = data_mask.to(device)
                label: torch.Tensor = label.to(device)
                pick_index: torch.Tensor = pick_index.to(device)
                label_mask: torch.Tensor = label_mask.to(device)

                # real training
                predicted_response, _ = self.gkt_model(question, data, data_mask)

                loss = loss_function(predicted_response, pick_index, label, label_mask)

                # back propagation
                trainer.zero_grad()
                loss.backward()
                trainer.step()

                losses.append(loss.mean().item())
            print("[Epoch %d] SLMoss: %.6f" % (e, float(np.mean(losses))))

            if test_data is not None:
                auc, accuracy = self.eval(test_data)
                print("[Epoch %d] auc: %.6f, accuracy: %.6f" % (e, auc, accuracy))

    def eval(self, test_data, device="cpu") -> tuple:
        self.gkt_model.eval()
        y_true = []
        y_pred = []

        for (question, data, data_mask, label, pick_index, label_mask) in tqdm(test_data, "evaluating"):
            # convert to device
            question: torch.Tensor = question.to(device)
            data: torch.Tensor = data.to(device)
            data_mask: torch.Tensor = data_mask.to(device)
            label: torch.Tensor = label.to(device)
            pick_index: torch.Tensor = pick_index.to(device)
            label_mask: torch.Tensor = label_mask.to(device)

            # real evaluating
            output, _ = self.gkt_model(question, data, data_mask)
            output = output[:, :-1]
            output = pick(output, pick_index.to(output.device))
            pred = tensor2list(output)
            label = tensor2list(label)
            for i, length in enumerate(label_mask.numpy().tolist()):
                length = int(length)
                y_true.extend(label[i][:length])
                y_pred.extend(pred[i][:length])
        self.gkt_model.train()
        return roc_auc_score(y_true, y_pred), accuracy_score(y_true, np.array(y_pred) >= 0.5)

    def save(self, filepath) -> ...:
        torch.save(self.gkt_model.state_dict(), filepath)
        logging.info("save parameters to %s" % filepath)

    def load(self, filepath):
        self.gkt_model.load_state_dict(torch.load(filepath))
        logging.info("load parameters from %s" % filepath)



# coding: utf-8
# 2021/5/26 @ tongshiwei

def pseudo_data_generation(ku_num, record_num=10, max_length=20):
    # 在这里定义测试用伪数据流
    import random
    random.seed(10)

    raw_data = [
        [
            (random.randint(0, ku_num - 1), random.randint(-1, 1))
            for _ in range(random.randint(2, max_length))
        ] for _ in range(record_num)
    ]

    return raw_data
